@@ -760,7 +760,12 @@ authorize_device_internal (BoltDevice *dev,
       int keyfd;
 
       bolt_debug (LOG_DEV (dev), LOG_TOPIC ("authorize"), "writing key");
-      keyfd = bolt_openat (dirfd (devdir), "key", O_WRONLY | O_CLOEXEC, error);
+
+      keyfd = bolt_openat (dirfd (devdir),
+                           "key",
+                           O_WRONLY | O_CLOEXEC,
+                           0,
+                           error);
       if (keyfd < 0)
         return FALSE;
 
@@ -1040,6 +1045,25 @@ handle_authorize (BoltExported          *object,
   BoltSecurity level;
   BoltKey *key;
 
+  /* In bolt_device_authorize the state is also checked, but it
+   * is done already here to fail quicker and avoid accessing the
+   * potentially unset domain (if e.g. the device is not connected).
+   */
+  if (!bolt_status_is_pending (dev->status))
+    {
+      g_set_error (error, BOLT_ERROR, BOLT_ERROR_BADSTATE,
+                   "wrong device state: %s",
+                   bolt_status_to_string (dev->status));
+      return NULL;
+    }
+  else if (dev->domain == NULL)
+    {
+      bolt_bug (LOG_DEV (dev), "device connected but no domain");
+      g_set_error_literal (error, BOLT_ERROR, BOLT_ERROR_BADSTATE,
+                           "device has no domain associated");
+      return NULL;
+    }
+
   level = bolt_domain_get_security (dev->domain);
   key = NULL;
 
@@ -1083,6 +1107,7 @@ bolt_device_new_for_udev (struct udev_device *udev,
 
   g_return_val_if_fail (udev != NULL, NULL);
   g_return_val_if_fail (domain != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
   uid = udev_device_get_sysattr_value (udev, "unique_id");
   if (uid == NULL)
@@ -1139,7 +1164,9 @@ bolt_device_export (BoltDevice      *device,
 {
   gboolean ok;
 
-  g_return_val_if_fail (BOLT_IS_DEVICE (device), FALSE);
+  g_return_val_if_fail (BOLT_IS_DEVICE (device), NULL);
+  g_return_val_if_fail (G_IS_DBUS_CONNECTION (connection), NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
   ok = bolt_exported_export (BOLT_EXPORTED (device),
                              connection,
@@ -1163,6 +1190,9 @@ bolt_device_authorize (BoltDevice         *dev,
 {
   g_autoptr(GTask) task = NULL;
 
+  g_return_if_fail (BOLT_IS_DEVICE (dev));
+  g_return_if_fail (BOLT_IS_AUTH (auth));
+
   task = authorize_prepare (dev, auth, callback, user_data);
 
   if (task == NULL)
@@ -1179,6 +1209,9 @@ bolt_device_authorize_idle (BoltDevice         *dev,
 {
   GTask *task = NULL;
 
+  g_return_if_fail (BOLT_IS_DEVICE (dev));
+  g_return_if_fail (BOLT_IS_AUTH (auth));
+
   task = authorize_prepare (dev, auth, callback, user_data);
 
   if (task == NULL)
@@ -1199,8 +1232,8 @@ bolt_device_connected (BoltDevice         *dev,
   gboolean ok;
   guint64 ct, at;
 
-  g_return_val_if_fail (dev != NULL, BOLT_STATUS_UNKNOWN);
-  g_return_val_if_fail (domain != NULL, BOLT_STATUS_UNKNOWN);
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), BOLT_STATUS_UNKNOWN);
+  g_return_val_if_fail (BOLT_IS_DOMAIN (domain), BOLT_STATUS_UNKNOWN);
 
   ok = bolt_sysfs_info_for_device (udev, TRUE, &info, &err);
   if (!ok)
@@ -1235,6 +1268,8 @@ bolt_device_connected (BoltDevice         *dev,
 BoltStatus
 bolt_device_disconnected (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), BOLT_STATUS_UNKNOWN);
+
   g_object_set (G_OBJECT (dev),
                 "parent", NULL,
                 "sysfs-path", NULL,
@@ -1253,14 +1288,18 @@ bolt_device_disconnected (BoltDevice *dev)
 }
 
 gboolean
-bolt_device_is_connected (const BoltDevice *device)
+bolt_device_is_connected (BoltDevice *device)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (device), FALSE);
+
   return bolt_status_is_connected (device->status);
 }
 
 gboolean
-bolt_device_is_authorized (const BoltDevice *device)
+bolt_device_is_authorized (BoltDevice *device)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (device), FALSE);
+
   return bolt_status_is_authorized (device->status);
 }
 
@@ -1276,6 +1315,9 @@ bolt_device_update_from_udev (BoltDevice         *dev,
   gboolean chg;
   gboolean ok;
 
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), BOLT_STATUS_UNKNOWN);
+  g_return_val_if_fail (udev != NULL, BOLT_STATUS_UNKNOWN);
+
   /* if we are currently authorizing, let's not update
    * the status, because we are most likely causing that
    * udev update and we cannot determine AUTHORIZING from
@@ -1322,21 +1364,25 @@ bolt_device_update_from_udev (BoltDevice         *dev,
 }
 
 BoltKeyState
-bolt_device_get_keystate (const BoltDevice *dev)
+bolt_device_get_keystate (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), BOLT_KEY_MISSING);
+
   return dev->key;
 }
 
 const char *
-bolt_device_get_name (const BoltDevice *dev)
+bolt_device_get_name (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), NULL);
+
   return dev->name;
 }
 
 const char *
 bolt_device_get_object_path (BoltDevice *device)
 {
-  g_return_val_if_fail (BOLT_IS_DEVICE (device), FALSE);
+  g_return_val_if_fail (BOLT_IS_DEVICE (device), NULL);
 
   if (!bolt_exported_is_exported (BOLT_EXPORTED (device)))
     return NULL;
@@ -1345,19 +1391,23 @@ bolt_device_get_object_path (BoltDevice *device)
 }
 
 BoltPolicy
-bolt_device_get_policy (const BoltDevice *dev)
+bolt_device_get_policy (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), BOLT_POLICY_UNKNOWN);
+
   return dev->policy;
 }
 
 const char *
-bolt_device_get_uid (const BoltDevice *dev)
+bolt_device_get_uid (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), NULL);
+
   return dev->uid;
 }
 
 BoltSecurity
-bolt_device_get_security (const BoltDevice *dev)
+bolt_device_get_security (BoltDevice *dev)
 {
   g_return_val_if_fail (dev != NULL, BOLT_SECURITY_UNKNOWN);
   g_return_val_if_fail (dev->domain != NULL, BOLT_SECURITY_UNKNOWN);
@@ -1366,93 +1416,131 @@ bolt_device_get_security (const BoltDevice *dev)
 }
 
 BoltStatus
-bolt_device_get_status (const BoltDevice *dev)
+bolt_device_get_status (BoltDevice *dev)
 {
+  g_return_val_if_fail (dev != NULL, BOLT_STATUS_UNKNOWN);
+
   return dev->status;
 }
 
 BoltAuthFlags
-bolt_device_get_authflags (const BoltDevice *dev)
+bolt_device_get_authflags (BoltDevice *dev)
 {
+  g_return_val_if_fail (dev != NULL, BOLT_AUTH_NONE);
+
   return dev->aflags;
 }
 
 gboolean
-bolt_device_get_stored (const BoltDevice *dev)
+bolt_device_get_stored (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), FALSE);
+
   return dev->store != NULL;
 }
 
 const char *
-bolt_device_get_syspath (const BoltDevice *dev)
+bolt_device_get_syspath (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), NULL);
+
   return dev->syspath;
 }
 
 const char *
-bolt_device_get_vendor (const BoltDevice *dev)
+bolt_device_get_vendor (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), NULL);
+
   return dev->vendor;
 }
 
 BoltDeviceType
-bolt_device_get_device_type (const BoltDevice *dev)
+bolt_device_get_device_type (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), BOLT_DEVICE_UNKNOWN_TYPE);
+
   return dev->type;
 }
 
 const char *
-bolt_device_get_label (const BoltDevice *dev)
+bolt_device_get_label (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), NULL);
+
   return dev->label;
 }
 
 guint64
-bolt_device_get_authtime (const BoltDevice *dev)
+bolt_device_get_authtime (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), 0);
+
   return dev->authtime;
 }
 
 guint64
-bolt_device_get_conntime (const BoltDevice *dev)
+bolt_device_get_conntime (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), 0);
+
   return dev->conntime;
 }
 
 gint64
-bolt_device_get_storetime (const BoltDevice *dev)
+bolt_device_get_storetime (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), 0);
+
   return dev->storetime;
 }
 
 gboolean
-bolt_device_supports_secure_mode (const BoltDevice *dev)
+bolt_device_supports_secure_mode (BoltDevice *dev)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), FALSE);
+
   return bolt_flag_isclear (dev->aflags, BOLT_AUTH_NOKEY);
 }
 
 gboolean
-bolt_device_check_authflag (const BoltDevice *dev,
-                            BoltAuthFlags     flag)
+bolt_device_check_authflag (BoltDevice   *dev,
+                            BoltAuthFlags flag)
 {
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), FALSE);
+
   return bolt_flag_isset (dev->aflags, flag);
 }
 
-BoltKey *
-bolt_device_get_key_from_sysfs (const BoltDevice *dev,
-                                GError          **error)
+gboolean
+bolt_device_get_key_from_sysfs (BoltDevice *dev,
+                                BoltKey   **key,
+                                GError    **error)
 {
+  g_autoptr(GError) err = NULL;
   g_autoptr(GFile) keyfile = NULL;
   g_autofree char *keypath = NULL;
-  BoltKey *key;
 
-  g_return_val_if_fail (dev != NULL, NULL);
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), FALSE);
+  g_return_val_if_fail (key != NULL && *key == NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
   if (dev->syspath == NULL)
-    return NULL;
+    {
+      g_set_error_literal (error, BOLT_ERROR, BOLT_ERROR_BADSTATE,
+                           "device is not connected");
+      return FALSE;
+    }
 
   keypath = g_build_filename (dev->syspath, "key", NULL);
   keyfile = g_file_new_for_path (keypath);
 
-  return key = bolt_key_load_file (keyfile, error);
+  *key = bolt_key_load_file (keyfile, &err);
+
+  if (*key != NULL)
+    return TRUE;
+  else if (bolt_err_notfound (err) || bolt_err_nokey (err))
+    return TRUE;
+  else
+    return bolt_error_propagate (error, &err);
 }
