@@ -488,7 +488,7 @@ dispach_property_setter (BoltExported          *exported,
 
   if (!ok && err != NULL)
     {
-      g_propagate_error (error, g_steal_pointer (&err));
+      bolt_error_propagate (error, &err);
       return NULL;
     }
   else if (!ok)
@@ -730,7 +730,7 @@ handle_dbus_get_property (GDBusConnection *connection,
   if (prop == NULL)
     {
       bolt_warn_err (err, LOG_TOPIC ("dbus"), "get_property");
-      g_propagate_error (error, g_steal_pointer (&err));
+      bolt_error_propagate (error, &err);
       return NULL;
     }
 
@@ -879,6 +879,12 @@ bolt_exported_class_set_interface_info (BoltExportedClass *klass,
   g_autoptr(GError) err = NULL;
   const char *xml;
 
+  g_return_if_fail (BOLT_IS_EXPORTED_CLASS (klass));
+  g_return_if_fail (klass->priv != NULL);
+  g_return_if_fail (klass->priv->node_info == NULL);
+  g_return_if_fail (iface_name != NULL);
+  g_return_if_fail (resource_name != NULL);
+
   bolt_exported_class_set_interface_name (klass, iface_name);
 
   data = g_resources_lookup_data (resource_name,
@@ -928,6 +934,8 @@ bolt_exported_class_export_property (BoltExportedClass *klass,
       return;
     }
 
+  g_return_if_fail (G_IS_PARAM_SPEC (spec));
+
   priv = klass->priv;
 
   name_obj = g_param_spec_get_name (spec);
@@ -992,6 +1000,7 @@ bolt_exported_class_export_properties (BoltExportedClass *klass,
                                        guint              n_pspecs,
                                        GParamSpec       **specs)
 {
+  g_return_if_fail (BOLT_IS_EXPORTED_CLASS (klass));
   g_return_if_fail (start > 0);
 
   for (guint i = start; i < n_pspecs; i++)
@@ -1031,6 +1040,10 @@ bolt_exported_class_export_method (BoltExportedClass        *klass,
 {
   BoltExportedMethod *method;
 
+  g_return_if_fail (BOLT_IS_EXPORTED_CLASS (klass));
+  g_return_if_fail (name != NULL);
+  g_return_if_fail (handler != NULL);
+
   method = g_new0 (BoltExportedMethod, 1);
 
   method->name = g_strdup (name);
@@ -1044,15 +1057,17 @@ bolt_exported_class_export_method (BoltExportedClass        *klass,
 gboolean
 bolt_exported_export (BoltExported    *exported,
                       GDBusConnection *connection,
-                      const char      *object_path,
+                      const char      *path_hint,
                       GError         **error)
 {
+  g_autofree char *object_path = NULL;
   BoltExportedPrivate *priv;
   BoltExportedClass *klass;
   guint id;
 
   g_return_val_if_fail (BOLT_IS_EXPORTED (exported), FALSE);
   g_return_val_if_fail (connection != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
   priv = GET_PRIV (exported);
   klass = BOLT_EXPORTED_GET_CLASS (exported);
@@ -1064,6 +1079,8 @@ bolt_exported_export (BoltExported    *exported,
       return FALSE;
     }
 
+  object_path = g_strdup (path_hint);
+
   if (object_path == NULL)
     {
       object_path = bolt_exported_make_object_path (exported);
@@ -1092,7 +1109,7 @@ bolt_exported_export (BoltExported    *exported,
   bolt_debug (LOG_TOPIC ("dbus"), "registered object at %s", object_path);
 
   priv->dbus = g_object_ref (connection);
-  priv->object_path = g_strdup (object_path);
+  priv->object_path = g_steal_pointer (&object_path);
   priv->registration = id;
 
   g_object_notify_by_pspec (G_OBJECT (exported), props[PROP_OBJECT_PATH]);
@@ -1179,6 +1196,9 @@ bolt_exported_emit_signal (BoltExported *exported,
   gboolean ok;
 
   g_return_val_if_fail (BOLT_IS_EXPORTED (exported), FALSE);
+  g_return_val_if_fail (name != NULL, FALSE);
+  g_return_val_if_fail (parameters != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
   priv = GET_PRIV (exported);
 
@@ -1200,7 +1220,7 @@ bolt_exported_emit_signal (BoltExported *exported,
     {
       bolt_warn_err (err, LOG_TOPIC ("dbus"),
                      "error emitting signal");
-      g_propagate_error (error, g_steal_pointer (&err));
+      bolt_error_propagate (error, &err);
     }
   else
     {
@@ -1229,8 +1249,6 @@ bolt_exported_prop_free (gpointer data)
   g_param_spec_unref (prop->spec);
   g_variant_type_free (prop->signature);
 
-
-
   g_free (prop);
 }
 
@@ -1238,26 +1256,23 @@ static GVariant *
 enum_gvalue_to_gvariant (GEnumClass   *enum_class,
                          const GValue *value)
 {
-  g_autofree char *str = NULL;
-  const char *name;
-  GEnumValue *ev;
+  g_autoptr(GError) err = NULL;
+  const char *str;
   GVariant *res;
   gint iv;
 
   iv = g_value_get_enum (value);
-  ev = g_enum_get_value (enum_class, iv);
+  str = bolt_enum_class_to_string (enum_class, iv, &err);
 
-  if (ev != NULL)
+  if (str == NULL)
     {
-      res = g_variant_new_string (ev->value_nick);
-      return g_variant_ref_sink (res);
+      bolt_bug ("auto-convert enum: %s", err->message);
+      res = g_variant_new_printf ("%d", iv);
+    }
+  else
+    {
+      res = g_variant_new_string (str);
     }
-
-  /* we got an invalid value for that enum */
-  str = g_strdup_printf ("%d", iv);
-  res = g_variant_new_string (str);
-  name = g_type_name_from_class ((GTypeClass *) enum_class);
-  bolt_bug ("invalid enum value %d for enum '%s'", iv, name);
 
   return g_variant_ref_sink (res);
 }
@@ -1268,28 +1283,18 @@ enum_gvariant_to_gvalue (GEnumClass *enum_class,
                          GValue     *value,
                          GError    **error)
 {
-  GEnumValue *ev = NULL;
   const char *str;
+  gboolean ok;
+  gint val;
 
   str = g_variant_get_string (variant, NULL);
 
-  if (str == NULL)
-    str = "invalid";
+  /* NB: it is ok to pass NULL here */
+  ok = bolt_enum_class_from_string (enum_class, str, &val, error);
 
-  ev = g_enum_get_value_by_nick (enum_class, str);
+  if (ok)
+    g_value_set_enum (value, val);
 
-  if (ev == NULL)
-    {
-      const char *name;
-
-      name = g_type_name_from_class ((GTypeClass *) enum_class);
-      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
-                   "invalid enum value '%s' for '%s'", str, name);
-
-      return FALSE;
-    }
-
-  g_value_set_enum (value, ev->value);
   return TRUE;
 }
 
@@ -1307,11 +1312,8 @@ flags_gvalue_to_gvariant (GFlagsClass  *flags_class,
   str = bolt_flags_class_to_string (flags_class, iv, &err);
   if (str == NULL)
     {
-      const char *name;
-
-      name = g_type_name_from_class ((GTypeClass *) flags_class);
+      bolt_bug ("auto-convert: %s", err->message);
       str = g_strdup_printf ("%u", iv);
-      bolt_bug ("invalid enum flags '%u' for enum '%s'", iv, name);
     }
 
   res = g_variant_new_string (str);
@@ -1324,21 +1326,13 @@ flags_gvariant_to_gvalue (GFlagsClass *flags_class,
                           GValue      *value,
                           GError     **error)
 {
-  gboolean ok;
   const char *str;
+  gboolean ok;
   guint flags;
 
   str = g_variant_get_string (variant, NULL);
 
-  if (str == NULL)
-    {
-      const char *name;
-
-      name = g_type_name_from_class ((GTypeClass *) flags_class);
-      g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
-                   "invalid flags value (null) for '%s'", name);
-    }
-
+  /* NB: NULL is a safe value for 'str' to be passed into */
   ok = bolt_flags_class_from_string (flags_class, str, &flags, error);
 
   if (ok)
