@@ -31,6 +31,7 @@
 #include "bolt-str.h"
 #include "bolt-sysfs.h"
 #include "bolt-udev.h"
+#include "bolt-unix.h"
 
 #include "bolt-manager.h"
 
@@ -50,7 +51,20 @@ static void     bolt_manager_initable_iface_init (GInitableIface *iface);
 static gboolean bolt_manager_initialize (GInitable    *initable,
                                          GCancellable *cancellable,
                                          GError      **error);
+
+/* internal manager functions */
+static void          manager_sd_notify_status (BoltManager *mgr);
+
 /* domain related functions */
+static gboolean      manager_load_domains (BoltManager *mgr,
+                                           GError     **error);
+
+static void          manager_bootacl_inital_sync (BoltManager *mgr,
+                                                  BoltDomain  *domain);
+
+static BoltDomain *  manager_domain_ensure (BoltManager        *mgr,
+                                            struct udev_device *dev);
+
 static BoltDomain *  manager_find_domain_by_syspath (BoltManager *mgr,
                                                      const char  *syspath);
 
@@ -61,6 +75,9 @@ static void          manager_deregister_domain (BoltManager *mgr,
                                                 BoltDomain  *domain);
 
 /* device related functions */
+static gboolean      manager_load_devices (BoltManager *mgr,
+                                           GError     **error);
+
 static void          manager_register_device (BoltManager *mgr,
                                               BoltDevice  *device);
 
@@ -93,9 +110,6 @@ static void          handle_udev_domain_event (BoltManager        *mgr,
                                                struct udev_device *device,
                                                const char         *action);
 
-static void          handle_udev_domain_added (BoltManager        *mgr,
-                                               struct udev_device *udev);
-
 static void          handle_udev_domain_removed (BoltManager *mgr,
                                                  BoltDomain  *domain);
 
@@ -104,6 +118,7 @@ static void          handle_udev_device_event (BoltManager        *mgr,
                                                const char         *action);
 
 static void          handle_udev_device_added (BoltManager        *mgr,
+                                               BoltDomain         *domain,
                                                struct udev_device *udev);
 
 static void          handle_udev_device_changed (BoltManager        *mgr,
@@ -114,6 +129,7 @@ static void          handle_udev_device_removed (BoltManager *mgr,
                                                  BoltDevice  *dev);
 
 static void          handle_udev_device_attached (BoltManager        *mgr,
+                                                  BoltDomain         *domain,
                                                   BoltDevice         *dev,
                                                   struct udev_device *udev);
 
@@ -121,10 +137,18 @@ static void          handle_udev_device_detached (BoltManager *mgr,
                                                   BoltDevice  *dev);
 
 /* signal callbacks */
+static void          handle_store_device_added (BoltStore   *store,
+                                                const char  *uid,
+                                                BoltManager *mgr);
+
 static void          handle_store_device_removed (BoltStore   *store,
                                                   const char  *uid,
                                                   BoltManager *mgr);
 
+static void          handle_domain_security_changed (BoltManager *mgr,
+                                                     GParamSpec  *unused,
+                                                     BoltDomain  *domain);
+
 static void          handle_device_status_changed (BoltDevice  *dev,
                                                    BoltStatus   old,
                                                    BoltManager *mgr);
@@ -256,11 +280,14 @@ bolt_manager_finalize (GObject *object)
       mgr->probing_timeout = 0;
     }
 
+  g_clear_pointer (&mgr->probing_roots, g_ptr_array_unref);
+
   g_clear_object (&mgr->store);
   g_ptr_array_free (mgr->devices, TRUE);
   bolt_domain_clear (&mgr->domains);
 
   g_clear_object (&mgr->power);
+  g_clear_object (&mgr->bouncer);
 
   G_OBJECT_CLASS (bolt_manager_parent_class)->finalize (object);
 }
@@ -320,6 +347,10 @@ bolt_manager_init (BoltManager *mgr)
   mgr->policy = BOLT_POLICY_AUTO;
   mgr->authmode = BOLT_AUTH_ENABLED;
 
+  g_signal_connect_object (mgr->store, "device-added",
+                           G_CALLBACK (handle_store_device_added),
+                           mgr, 0);
+
   g_signal_connect_object (mgr->store, "device-removed",
                            G_CALLBACK (handle_store_device_removed),
                            mgr, 0);
@@ -426,11 +457,11 @@ bolt_manager_initialize (GInitable    *initable,
                          GCancellable *cancellable,
                          GError      **error)
 {
-  g_auto(GStrv) ids = NULL;
   g_autoptr(BoltPowerGuard) power = NULL;
   BoltManager *mgr;
   struct udev_enumerate *enumerate;
   struct udev_list_entry *l, *devices;
+  gboolean ok;
 
   mgr = BOLT_MANAGER (initable);
 
@@ -455,33 +486,13 @@ bolt_manager_initialize (GInitable    *initable,
                            (GCallback) handle_uevent_udev,
                            mgr, 0);
 
-  ids = bolt_store_list_uids (mgr->store, error);
-  if (ids == NULL)
-    {
-      g_prefix_error (error, "failed to list devices in store");
-      return FALSE;
-    }
+  ok = manager_load_domains (mgr, error);
+  if (!ok)
+    return FALSE;
 
-  bolt_info (LOG_TOPIC ("store"), "loading devices");
-  for (guint i = 0; i < g_strv_length (ids); i++)
-    {
-      g_autoptr(GError) err = NULL;
-      BoltDevice *dev = NULL;
-      const char *uid = ids[i];
-
-      bolt_info (LOG_DEV_UID (uid), LOG_TOPIC ("store"), "loading device");
-
-      dev = bolt_store_get_device (mgr->store, uid, &err);
-      if (dev == NULL)
-        {
-          bolt_warn_err (err, LOG_TOPIC ("store"),
-                         LOG_DIRECT (BOLT_LOG_DEVICE_UID, uid),
-                         "failed to load device (%.7s)", uid);
-          continue;
-        }
-
-      manager_register_device (mgr, dev);
-    }
+  ok = manager_load_devices (mgr, error);
+  if (!ok)
+    return FALSE;
 
   /* setup the power controller */
   mgr->power = bolt_power_new (mgr->udev);
@@ -511,8 +522,6 @@ bolt_manager_initialize (GInitable    *initable,
     {
       g_autoptr(GError) err = NULL;
       g_autoptr(udev_device) udevice = NULL;
-      g_autoptr(BoltDevice) dev = NULL;
-      const char *uid;
       const char *syspath;
       const char *devtype;
 
@@ -530,31 +539,275 @@ bolt_manager_initialize (GInitable    *initable,
       devtype = udev_device_get_devtype (udevice);
 
       if (bolt_streq (devtype, "thunderbolt_domain"))
-        handle_udev_domain_added (mgr, udevice);
+        handle_udev_domain_event (mgr, udevice, "add");
 
       if (!bolt_streq (devtype, "thunderbolt_device"))
         continue;
 
-      uid = udev_device_get_sysattr_value (udevice, "unique_id");
-      if (uid == NULL)
-        {
-          bolt_warn ("thunderbolt device without uid");
-          continue;
-        }
-
-      dev = manager_find_device_by_uid (mgr, uid, NULL);
-      if (dev)
-        handle_udev_device_attached (mgr, dev, udevice);
-      else
-        handle_udev_device_added (mgr, udevice);
+      handle_udev_device_event (mgr, udevice, "add");
     }
 
   udev_enumerate_unref (enumerate);
 
+  manager_sd_notify_status (mgr);
+
   return TRUE;
 }
 
+/* internal functions */
+static void
+manager_sd_notify_status (BoltManager *mgr)
+{
+  g_autoptr(GError) err = NULL;
+  g_autofree char *status = NULL;
+  const char *pstate = NULL;
+  BoltPowerState power;
+  gboolean sent;
+  gboolean authorizing;
+  gboolean ok = FALSE;
+
+  authorizing = bolt_flag_isset (mgr->authmode, BOLT_AUTH_ENABLED);
+  power = bolt_power_get_state (mgr->power);
+  pstate = bolt_enum_to_string (BOLT_TYPE_POWER_STATE, power, NULL);
+
+  status = g_strdup_printf ("STATUS=authmode: %s,"
+                            " force-power: %s",
+                            (authorizing ? "enabled" : "DISABLED"),
+                            (pstate ? : "unknown"));
+
+  ok = bolt_sd_notify_literal (status, &sent, &err);
+
+  if (!ok)
+    bolt_warn_err (err, LOG_TOPIC ("status"),
+                   "failed to send status");
+  else
+    bolt_debug (LOG_TOPIC ("status"), "%s [sent: %s]",
+                status, bolt_yesno (sent));
+}
+
+static void
+manager_maybe_set_security (BoltManager *mgr,
+                            BoltSecurity security)
+{
+  /* update the security level, if it is not already
+  * set, but also ignore if security is 'unknown' */
+
+  if (security == BOLT_SECURITY_UNKNOWN)
+    return;
+
+  if (mgr->security == BOLT_SECURITY_UNKNOWN)
+    {
+      bolt_info ("security level set to '%s'",
+                 bolt_security_to_string (security));
+      mgr->security = security;
+      g_object_notify_by_pspec (G_OBJECT (mgr),
+                                props[PROP_SECURITY]);
+    }
+  else if (mgr->security != security)
+    {
+      bolt_warn ("multiple security levels (%s vs %s)",
+                 bolt_security_to_string (mgr->security),
+                 bolt_security_to_string (security));
+    }
+}
+
 /* domain related function */
+static gboolean
+manager_load_domains (BoltManager *mgr,
+                      GError     **error)
+{
+  g_auto(GStrv) ids = NULL;
+
+  ids = bolt_store_list_uids (mgr->store, "domains", error);
+  if (ids == NULL)
+    {
+      g_prefix_error (error, "failed to list domains in store: ");
+      return FALSE;
+    }
+
+  bolt_info (LOG_TOPIC ("store"), "loading domains");
+
+  for (guint i = 0; i < g_strv_length (ids); i++)
+    {
+      g_autoptr(GError) err = NULL;
+      BoltDomain *dom = NULL;
+      const char *uid = ids[i];
+
+      bolt_info (LOG_TOPIC ("store"), LOG_DOM_UID (uid),
+                 "loading domain");
+
+      dom = bolt_store_get_domain (mgr->store, uid, &err);
+      if (dom == NULL)
+        {
+          bolt_warn_err (err, LOG_DOM_UID (uid), LOG_TOPIC ("store"),
+                         "failed to load domain", uid);
+          continue;
+        }
+
+      manager_register_domain (mgr, dom);
+    }
+
+  return TRUE;
+}
+
+static void
+manager_bootacl_inital_sync (BoltManager *mgr,
+                             BoltDomain  *domain)
+{
+  g_autoptr(GError) err = NULL;
+  g_auto(GStrv) acl = NULL;
+  gboolean ok;
+  guint n, empty;
+
+  if (!bolt_domain_supports_bootacl (domain))
+    {
+      bolt_info (LOG_TOPIC ("bootacl"), LOG_DOM (domain),
+                 "bootacl not supported, no sync");
+      return;
+    }
+
+  acl = bolt_domain_dup_bootacl (domain);
+  g_assert (acl != NULL);
+
+  n = bolt_domain_bootacl_slots (domain, &empty);
+
+  bolt_info (LOG_TOPIC ("bootacl"), LOG_DOM (domain),
+             "sync start [slots: %u free: %u]", n, empty);
+
+  for (guint i = 0; i < mgr->devices->len; i++)
+    {
+      BoltDevice *dev = g_ptr_array_index (mgr->devices, i);
+      const char *duid = bolt_device_get_uid (dev);
+      gboolean polok, inacl, sync;
+
+      polok = bolt_device_get_policy (dev) == BOLT_POLICY_AUTO;
+      inacl = bolt_domain_bootacl_contains (domain, duid);
+      sync = polok && !inacl;
+
+      bolt_info (LOG_TOPIC ("bootacl"), LOG_DOM (domain),
+                 LOG_DEV_UID (duid),
+                 "sync '%.13s…' %s [policy: %3s, in acl: %3s]",
+                 duid, bolt_yesno (sync), bolt_yesno (polok),
+                 bolt_yesno (inacl));
+
+      if (!sync)
+        continue;
+
+      bolt_domain_bootacl_allocate (domain, acl, duid);
+    }
+
+  ok = bolt_domain_bootacl_set (domain, acl, &err);
+
+  if (!ok && err != NULL)
+    {
+      bolt_warn_err (err, LOG_DOM (domain), "failed to write bootacl");
+      return;
+    }
+
+  bolt_domain_bootacl_slots (domain, &empty);
+  bolt_info (LOG_TOPIC ("bootacl"), LOG_DOM (domain),
+             "sync done [wrote: %s, now free: %u]",
+             bolt_yesno (ok), empty);
+}
+
+static BoltDomain *
+manager_domain_ensure (BoltManager        *mgr,
+                       struct udev_device *dev)
+{
+  g_autoptr(GError) err = NULL;
+  BoltSecurity security;
+  BoltDomain *domain = NULL;
+  GDBusConnection *bus;
+  struct udev_device *host;
+  struct udev_device *dom;
+  const char *syspath;
+  const char *op;
+  const char *uid;
+  gboolean ok;
+
+  /* check if we already know a domain that is the parent
+   * of the device (dev); if not then 'dev' is very likely
+   * the host device.
+   */
+  syspath = udev_device_get_syspath (dev);
+  domain = manager_find_domain_by_syspath (mgr, syspath);
+
+  if (domain != NULL)
+    return domain;
+
+  /* dev is very likely the host, i.e. root switch device,
+   * but we might as well make sure we can get the domain
+   * from any; also we make sure that we have indeed the
+   * host device via this lookup.
+   */
+  dom = bolt_sysfs_domain_for_device (dev, &host);
+  if (dom == NULL)
+    return NULL;
+
+  uid = udev_device_get_sysattr_value (host, "unique_id");
+
+  /* check if we have a stored domain with a matching uid
+   * of the host device, if so then we have just connected
+   * the corresponding domain controller, represented by
+   * the 'dom' udev_device.
+   */
+  domain = bolt_domain_find_id (mgr->domains, uid, NULL);
+
+  if (domain != NULL)
+    {
+      bolt_domain_connected (domain, dom);
+      return domain;
+    }
+
+  /* this is an unknown, unstored domain controller */
+  domain = bolt_domain_new_for_udev (dom, uid, &err);
+
+  if (domain == NULL)
+    {
+      bolt_warn_err (err, LOG_TOPIC ("udev"),
+                     "failed to create domain: %s");
+      return NULL;
+    }
+
+  security = bolt_domain_get_security (domain);
+
+  bolt_msg (LOG_DOM (domain), "newly connected [%s] (%s)",
+            bolt_security_to_string (security), syspath);
+
+  manager_maybe_set_security (mgr, security);
+
+  manager_register_domain (mgr, domain);
+
+  /* registering the domain will add one reference */
+  g_object_unref (domain);
+
+  /* add all devices with POLICY_AUTO to the bootacl */
+  manager_bootacl_inital_sync (mgr, domain);
+
+  /* now store the domain (with an updated bootacl) */
+  bolt_info (LOG_TOPIC ("store"), LOG_DOM (domain),
+             "storing newly connected domain");
+
+  ok = bolt_store_put_domain (mgr->store, domain, &err);
+  if (!ok)
+    bolt_warn_err (err, LOG_TOPIC ("store"), LOG_DOM (domain),
+                   "could not store domain");
+
+  bus = bolt_exported_get_connection (BOLT_EXPORTED (mgr));
+  if (bus == NULL)
+    return domain;
+
+  bolt_domain_export (domain, bus);
+
+  op = bolt_exported_get_object_path (BOLT_EXPORTED (domain));
+  bolt_exported_emit_signal (BOLT_EXPORTED (mgr),
+                             "DomainAdded",
+                             g_variant_new ("(o)", op),
+                             NULL);
+
+  return domain;
+}
+
 static BoltDomain *
 manager_find_domain_by_syspath (BoltManager *mgr,
                                 const char  *syspath)
@@ -570,7 +823,7 @@ manager_find_domain_by_syspath (BoltManager *mgr,
       /* we get a perfect match, if we search for the domain
        * itself, or if we are looking for the domain that
        * is the parent of the device in @syspath */
-      if (g_str_has_prefix (syspath, prefix))
+      if (prefix && g_str_has_prefix (syspath, prefix))
         return iter;
 
       iter = bolt_domain_next (iter);
@@ -579,35 +832,74 @@ manager_find_domain_by_syspath (BoltManager *mgr,
   return NULL;
 }
 
+static gboolean
+bootacl_alloc (BoltDomain  *domain,
+               GStrv        acl,
+               const char  *uid,
+               gint        *slot,
+               BoltManager *mgr)
+{
+  char **target = NULL;
+  guint64 last = G_MAXUINT64;
+
+  /* if we have a valid allocation, just use that */
+  if (*slot != -1)
+    return TRUE;
+
+  g_return_val_if_fail (acl != NULL, FALSE);
+
+  /* we need to replace one of devices currently in the acl */
+  for (char **iter = acl; *iter; iter++)
+    {
+      g_autoptr(BoltDevice) dev = NULL;
+      guint64 ts;
+
+      dev = manager_find_device_by_uid (mgr, uid, NULL);
+
+      /* device might not be managed by us */
+      if (dev == NULL)
+        continue;
+
+      /* authtime is the last time a device was seen and
+       * successfully authorized; choose the one that
+       * has the *oldest* timestamp (smallest number) */
+      ts = bolt_device_get_authtime (dev);
+      if (ts < last)
+        {
+          target = iter;
+          last = ts;
+        }
+    }
+
+  if (target == NULL)
+    return FALSE;
+
+  *slot = target - acl;
+
+  return TRUE;
+}
+
+
 static void
 manager_register_domain (BoltManager *mgr,
                          BoltDomain  *domain)
 {
-  const char *name;
-  BoltSecurity sl;
+  guint n_slots, n_free;
 
   mgr->domains = bolt_domain_insert (mgr->domains, domain);
 
-  name = bolt_domain_get_id (domain);
-  sl = bolt_domain_get_security (domain);
+  n_slots = bolt_domain_bootacl_slots (domain, &n_free);
 
-  bolt_info (LOG_TOPIC ("domain"), "'%s' (security: %s) added",
-             name, bolt_security_to_string (sl));
+  bolt_info (LOG_TOPIC ("domain"), LOG_DOM (domain),
+             "registered (bootacl: %u/%u)",
+             n_free, n_slots);
 
-  if (mgr->security == BOLT_SECURITY_UNKNOWN)
-    {
-      bolt_info ("security level set to '%s'",
-                 bolt_security_to_string (sl));
-      mgr->security = sl;
-      g_object_notify_by_pspec (G_OBJECT (mgr),
-                                props[PROP_SECURITY]);
-    }
-  else if (mgr->security != sl)
-    {
-      bolt_warn ("multiple security levels (%s vs %s)",
-                 bolt_security_to_string (mgr->security),
-                 bolt_security_to_string (sl));
-    }
+  g_signal_connect_object (domain, "bootacl-alloc",
+                           G_CALLBACK (bootacl_alloc), mgr, 0);
+
+  g_signal_connect_object (domain, "notify::security",
+                           G_CALLBACK (handle_domain_security_changed),
+                           mgr, G_CONNECT_SWAPPED);
 }
 
 static void
@@ -623,6 +915,43 @@ manager_deregister_domain (BoltManager *mgr,
 }
 
 /* device related functions */
+static gboolean
+manager_load_devices (BoltManager *mgr,
+                      GError     **error)
+{
+  g_auto(GStrv) ids = NULL;
+
+  ids = bolt_store_list_uids (mgr->store, "devices", error);
+  if (ids == NULL)
+    {
+      g_prefix_error (error, "failed to list devices in store: ");
+      return FALSE;
+    }
+
+  bolt_info (LOG_TOPIC ("store"), "loading devices");
+  for (guint i = 0; i < g_strv_length (ids); i++)
+    {
+      g_autoptr(GError) err = NULL;
+      BoltDevice *dev = NULL;
+      const char *uid = ids[i];
+
+      bolt_info (LOG_DEV_UID (uid), LOG_TOPIC ("store"), "loading device");
+
+      dev = bolt_store_get_device (mgr->store, uid, &err);
+      if (dev == NULL)
+        {
+          bolt_warn_err (err, LOG_TOPIC ("store"),
+                         LOG_DIRECT (BOLT_LOG_DEVICE_UID, uid),
+                         "failed to load device (%.7s)", uid);
+          continue;
+        }
+
+      manager_register_device (mgr, dev);
+    }
+
+  return TRUE;
+}
+
 static void
 manager_register_device (BoltManager *mgr,
                          BoltDevice  *dev)
@@ -891,8 +1220,8 @@ manager_maybe_auto_import_device (BoltManager *mgr,
       return;
     }
 
-  key = bolt_device_get_key_from_sysfs (dev, &err);
-  if (key == NULL && !bolt_err_notfound (err))
+  ok = bolt_device_get_key_from_sysfs (dev, &key, &err);
+  if (!ok)
     {
       bolt_warn_err (err, LOG_DEV (dev), LOG_TOPIC ("auto-import"),
                      "failed to read key from sysfs");
@@ -965,64 +1294,50 @@ handle_udev_domain_event (BoltManager        *mgr,
 
   syspath = udev_device_get_syspath (device);
 
-  if (g_str_equal (action, "add") ||
-      g_str_equal (action, "change"))
+  if (g_str_equal (action, "add"))
+    {
+      manager_probing_domain_added (mgr, device);
+
+      /* the creation of the actual domain object and
+       * its registration is handled on-demand: only
+       * when the host device appears, the uevent
+       * handler for the thunderbolt devices will
+       * ensure the domain exists
+       */
+    }
+  else if (g_str_equal (action, "change"))
     {
       domain = manager_find_domain_by_syspath (mgr, syspath);
 
-      if (domain != NULL)
-        return; /*change event, ignore for now */
+      if (domain == NULL)
+        {
+          bolt_warn (LOG_TOPIC ("domain"),
+                     "unregistered domain changed at %s",
+                     syspath);
+          return;
+        }
 
-      handle_udev_domain_added (mgr, device);
+      bolt_domain_update_from_udev (domain, device);
     }
   else if (g_str_equal (action, "remove"))
     {
       domain = manager_find_domain_by_syspath (mgr, syspath);
 
-      if (domain)
-        handle_udev_domain_removed (mgr, domain);
+      if (!domain)
+        {
+          bolt_warn (LOG_TOPIC ("domain"),
+                     "unregistered domain removed at %s",
+                     syspath);
+          return;
+        }
+
+      if (bolt_domain_is_stored (domain))
+        bolt_domain_disconnected (domain);
       else
-        bolt_warn (LOG_TOPIC ("domain"),
-                   "unregistered domain removed at %s",
-                   syspath);
+        handle_udev_domain_removed (mgr, domain);
     }
 }
 
-static void
-handle_udev_domain_added (BoltManager        *mgr,
-                          struct udev_device *device)
-{
-  g_autoptr(GError) err = NULL;
-  g_autoptr(BoltDomain) domain = NULL;
-  GDBusConnection *bus;
-  const char *op;
-
-  manager_probing_domain_added (mgr, device);
-
-  domain = bolt_domain_new_for_udev (device, &err);
-
-  if (domain == NULL)
-    {
-      bolt_warn_err (err, LOG_TOPIC ("udev"),
-                     "failed to create domain: %s");
-      return;
-    }
-
-  manager_register_domain (mgr, domain);
-
-  bus = bolt_exported_get_connection (BOLT_EXPORTED (mgr));
-  if (bus == NULL)
-    return;
-
-  bolt_domain_export (domain, bus);
-
-  op = bolt_exported_get_object_path (BOLT_EXPORTED (domain));
-  bolt_exported_emit_signal (BOLT_EXPORTED (mgr),
-                             "DomainAdded",
-                             g_variant_new ("(o)", op),
-                             NULL);
-}
-
 static void
 handle_udev_domain_removed (BoltManager *mgr,
                             BoltDomain  *domain)
@@ -1058,10 +1373,14 @@ handle_udev_device_event (BoltManager        *mgr,
                           const char         *action)
 {
   g_autoptr(BoltDevice) dev = NULL;
+  const char *syspath;
+
+  syspath = udev_device_get_syspath (device);
 
   if (g_str_equal (action, "add") ||
       g_str_equal (action, "change"))
     {
+      BoltDomain *dom;
       const char *uid;
 
       /* filter sysfs devices (e.g. the domain) that don't have
@@ -1070,27 +1389,29 @@ handle_udev_device_event (BoltManager        *mgr,
       if (uid == NULL)
         return;
 
+      dom = manager_domain_ensure (mgr, device);
+
+      if (dom == NULL)
+        {
+          bolt_warn (LOG_TOPIC ("domain"),
+                     "could not find domain for device at '%s'",
+                     syspath);
+          return;
+        }
+
       dev = manager_find_device_by_uid (mgr, uid, NULL);
 
       if (!dev)
-        handle_udev_device_added (mgr, device);
+        handle_udev_device_added (mgr, dom, device);
       else if (!bolt_device_is_connected (dev))
-        handle_udev_device_attached (mgr, dev, device);
+        handle_udev_device_attached (mgr, dom, dev, device);
       else
         handle_udev_device_changed (mgr, dev, device);
     }
   else if (g_str_equal (action, "remove"))
     {
-      const char *syspath;
       const char *name;
 
-      syspath = udev_device_get_syspath (device);
-      if (syspath == NULL)
-        {
-          bolt_warn (LOG_TOPIC ("udev"), "device without syspath");
-          return;
-        }
-
       /* filter out the domain controller */
       name = udev_device_get_sysname (device);
       if (name && g_str_has_prefix (name, "domain"))
@@ -1112,26 +1433,17 @@ handle_udev_device_event (BoltManager        *mgr,
 
 static void
 handle_udev_device_added (BoltManager        *mgr,
+                          BoltDomain         *domain,
                           struct udev_device *udev)
 {
   g_autoptr(GError) err = NULL;
   GDBusConnection *bus;
   BoltDevice *dev;
   BoltStatus status;
-  BoltDomain *domain;
   const char *opath;
   const char *syspath;
 
   syspath = udev_device_get_syspath (udev);
-  domain = manager_find_domain_by_syspath (mgr, syspath);
-
-  if (domain == NULL)
-    {
-      bolt_warn (LOG_TOPIC ("domain"),
-                 "could not find domain for device at '%s'",
-                 syspath);
-      return;
-    }
 
   dev = bolt_device_new_for_udev (udev, domain, &err);
   if (dev == NULL)
@@ -1218,25 +1530,15 @@ handle_udev_device_removed (BoltManager *mgr,
 
 static void
 handle_udev_device_attached (BoltManager        *mgr,
+                             BoltDomain         *domain,
                              BoltDevice         *dev,
                              struct udev_device *udev)
 {
   g_autoptr(BoltDevice) parent = NULL;
-  BoltDomain *domain;
   const char *syspath;
   BoltStatus status;
 
   syspath = udev_device_get_syspath (udev);
-  domain = manager_find_domain_by_syspath (mgr, syspath);
-
-  if (domain == NULL)
-    {
-      bolt_warn (LOG_TOPIC ("domain"),
-                 "could not find domain for device at '%s'",
-                 syspath);
-      return;
-    }
-
   status = bolt_device_connected (dev, domain, udev);
 
   bolt_msg (LOG_DEV (dev), "connected: %s (%s)",
@@ -1276,21 +1578,114 @@ handle_udev_device_detached (BoltManager *mgr,
   bolt_device_disconnected (dev);
 }
 
+typedef struct
+{
+  BoltManager *mgr;
+  const char  *uid;
+} BootaclCtx;
+
+static void
+bootacl_add_dev (gpointer data,
+                 gpointer user_data)
+{
+  g_autoptr(GError) err = NULL;
+  BoltDomain *dom = data;
+  BootaclCtx *ctx = user_data;
+  gboolean ok;
+
+  bolt_info (LOG_TOPIC ("bootacl"), LOG_DOM (dom),
+             LOG_DEV_UID (ctx->uid),
+             "adding newly stored device to bootacl");
+
+  if (!bolt_domain_supports_bootacl (dom))
+    return;
+
+  if (bolt_domain_bootacl_contains (dom, ctx->uid))
+    return;
+
+  ok = bolt_domain_bootacl_add (dom, ctx->uid, &err);
+  if (!ok)
+    {
+      bolt_warn_err (err, LOG_TOPIC ("bootacl"), LOG_DOM (dom),
+                     LOG_DEV_UID (ctx->uid),
+                     "could not add device [%.17s]",
+                     ctx->uid);
+    }
+}
+
+static void
+bootacl_del_dev (gpointer data,
+                 gpointer user_data)
+{
+  g_autoptr(GError) err = NULL;
+  BoltDomain *dom = data;
+  BootaclCtx *ctx = user_data;
+  gboolean ok;
+
+  bolt_info (LOG_TOPIC ("bootacl"), LOG_DOM (dom),
+             LOG_DEV_UID (ctx->uid),
+             "removing forgotten device from bootacl");
+
+  if (!bolt_domain_supports_bootacl (dom))
+    return;
+
+  if (!bolt_domain_bootacl_contains (dom, ctx->uid))
+    return;
+
+  ok = bolt_domain_bootacl_del (dom, ctx->uid, &err);
+  if (!ok)
+    {
+      bolt_warn_err (err, LOG_TOPIC ("bootacl"),
+                     LOG_DEV_UID (ctx->uid), LOG_DOM (dom),
+                     "could not remove device [%.17s]",
+                     ctx->uid);
+    }
+}
+
+
+static void
+handle_store_device_added (BoltStore   *store,
+                           const char  *uid,
+                           BoltManager *mgr)
+{
+  g_autoptr(BoltDevice) dev = NULL;
+  BoltDomain *dom = mgr->domains;
+  BootaclCtx ctx = {mgr, uid};
+
+  dev = manager_find_device_by_uid (mgr, uid, NULL);
+
+  if (dev == NULL || dom == NULL)
+    return;
+
+  if (bolt_device_get_policy (dev) != BOLT_POLICY_AUTO)
+    {
+      bolt_info (LOG_TOPIC ("bootacl"),
+                 LOG_DEV_UID (uid),
+                 "policy not 'auto', not adding");
+      return;
+    }
+
+  bolt_domain_foreach (dom, bootacl_add_dev, &ctx);
+}
+
 static void
 handle_store_device_removed (BoltStore   *store,
                              const char  *uid,
                              BoltManager *mgr)
 {
   g_autoptr(BoltDevice) dev = NULL;
+  BoltDomain *dom = mgr->domains;
+  BootaclCtx ctx = {mgr, uid};
   BoltStatus status;
   const char *opath;
 
   dev = manager_find_device_by_uid (mgr, uid, NULL);
-  bolt_msg (LOG_DEV (dev), "removed from store");
 
   if (!dev)
     return;
 
+  bolt_msg (LOG_DEV (dev), "removed from store");
+
   /* TODO: maybe move to a new bolt_device_removed (dev) */
   g_object_set (dev,
                 "store", NULL,
@@ -1298,6 +1693,9 @@ handle_store_device_removed (BoltStore   *store,
                 "policy", BOLT_POLICY_DEFAULT,
                 NULL);
 
+  /* remove device from bootacl */
+  bolt_domain_foreach (dom, bootacl_del_dev, &ctx);
+
   status = bolt_device_get_status (dev);
   /* if the device is connected, keep it around */
   if (status != BOLT_STATUS_DISCONNECTED)
@@ -1319,6 +1717,19 @@ handle_store_device_removed (BoltStore   *store,
 }
 
 
+static void
+handle_domain_security_changed (BoltManager *mgr,
+                                GParamSpec  *unused,
+                                BoltDomain  *domain)
+{
+  BoltSecurity security = bolt_domain_get_security (domain);
+  gboolean online = bolt_domain_is_connected (domain);
+
+  if (online)
+    manager_maybe_set_security (mgr, security);
+}
+
+
 static void
 handle_device_status_changed (BoltDevice  *dev,
                               BoltStatus   old,
@@ -1364,8 +1775,16 @@ handle_power_state_changed (GObject    *gobject,
                             gpointer    user_data)
 {
   BoltManager *mgr = BOLT_MANAGER (user_data);
+  gboolean supported = bolt_power_can_force (mgr->power);
+  BoltPowerState state = bolt_power_get_state (mgr->power);
+
+  bolt_info (LOG_TOPIC ("power"), "state changed: %s/%s",
+             (supported ? "supported" : "unsupported"),
+             bolt_power_state_to_string (state));
 
   g_object_notify_by_pspec (G_OBJECT (mgr), props[PROP_POWERSTATE]);
+
+  manager_sd_notify_status (mgr);
 }
 
 
@@ -1660,8 +2079,7 @@ handle_set_authmode (BoltExported *obj,
   if (str == NULL)
     {
       bolt_warn_err (err, LOG_TOPIC ("config"), "error setting authmode");
-      g_propagate_error (error, g_steal_pointer (&err));
-      return FALSE;
+      return bolt_error_propagate (error, &err);
     }
 
   bolt_config_set_auth_mode (mgr->config, str);
@@ -1670,12 +2088,14 @@ handle_set_authmode (BoltExported *obj,
   if (!ok)
     {
       bolt_warn_err (err, LOG_TOPIC ("config"), "error saving config");
-      g_propagate_error (error, g_steal_pointer (&err));
-      return FALSE;
+      return bolt_error_propagate (error, &err);
     }
 
   mgr->authmode = authmode;
   bolt_info (LOG_TOPIC ("config"), "auth mode set to '%s'", str);
+
+  manager_sd_notify_status (mgr);
+
   return ok;
 }
 
@@ -1842,15 +2262,16 @@ enroll_device_store_authorized (BoltManager *mgr,
   const char *opath;
   gboolean ok;
 
-  bolt_info (LOG_DEV (dev), "enrolling an authorized device");
+  bolt_info (LOG_DEV (dev), "enrolling an authorized device (%s)",
+             bolt_policy_to_string (policy));
 
-  key = bolt_device_get_key_from_sysfs (dev, &err);
-  if (key == NULL && !bolt_err_notfound (err))
+  ok = bolt_device_get_key_from_sysfs (dev, &key, &err);
+  if (!ok)
     {
       bolt_warn_err (err, LOG_DEV (dev), LOG_TOPIC ("udev"),
                      "failed to read key from sysfs");
       g_propagate_prefixed_error (error, g_steal_pointer (&err), "%s",
-                                  "could not determine existing authorization");
+                                  "could not determine existing authorization: ");
       return NULL;
     }
 
@@ -1860,7 +2281,7 @@ enroll_device_store_authorized (BoltManager *mgr,
     {
       bolt_warn_err (err, LOG_DEV (dev), LOG_TOPIC ("store"),
                      "failed to store device");
-      g_propagate_error (error, g_steal_pointer (&err));
+      bolt_error_propagate (error, &err);
       return NULL;
     }
 
@@ -1900,6 +2321,11 @@ handle_enroll_device (BoltExported          *obj,
                      "invalid policy: %s", policy);
       return NULL;
     }
+  else if (pol == BOLT_POLICY_DEFAULT)
+    {
+      pol = mgr->policy;
+      bolt_info (LOG_DEV (dev), "got 'default' policy, adjusting");
+    }
 
   if (bolt_device_get_stored (dev))
     {
