@@ -127,11 +127,11 @@ bolt_power_guard_get_property (GObject    *object,
       break;
 
     case PROP_PATH:
-      g_value_set_object (value, guard->path);
+      g_value_set_string (value, guard->path);
       break;
 
     case PROP_FIFO:
-      g_value_set_object (value, guard->fifo);
+      g_value_set_string (value, guard->fifo);
       break;
 
     case PROP_ID:
@@ -436,10 +436,7 @@ bolt_power_guard_mkfifo (BoltPowerGuard *guard,
 
   r = bolt_mkfifo (guard->fifo, 0600, &err);
   if (r == -1 && !bolt_err_exists (err))
-    {
-      g_propagate_error (error, g_steal_pointer (&err));
-      return FALSE;
-    }
+    return bolt_error_propagate (error, &err);
 
   g_object_notify_by_pspec (G_OBJECT (guard),
                             guard_props[PROP_FIFO]);
@@ -488,9 +485,12 @@ bolt_power_guard_monitor (BoltPowerGuard *guard,
   gboolean ok;
   int fd;
 
+  g_return_val_if_fail (BOLT_IS_POWER_GUARD (guard), -1);
+  g_return_val_if_fail (error == NULL || *error == NULL, -1);
+
   ok = bolt_power_guard_mkfifo (guard, error);
   if (!ok)
-    return FALSE;
+    return -1;
 
   /* reader */
   fd = bolt_open (guard->fifo, O_RDONLY | O_CLOEXEC | O_NONBLOCK, 0, error);
@@ -757,9 +757,9 @@ bolt_power_class_init (BoltPowerClass *klass)
                          G_PARAM_STATIC_STRINGS);
 
   power_props[PROP_STATEDIR] =
-    g_param_spec_string ("statedir",
+    g_param_spec_object ("statedir",
                          NULL, NULL,
-                         DEFAULT_RUNDIR,
+                         G_TYPE_FILE,
                          G_PARAM_READABLE |
                          G_PARAM_STATIC_STRINGS);
 
@@ -831,8 +831,6 @@ bolt_power_initialize (GInitable    *initable,
   g_autoptr(BoltPowerGuard) guard = NULL;
   g_autofree char *statedir = NULL;
   BoltPower *power = BOLT_POWER (initable);
-  struct udev_enumerate *e;
-  struct udev_list_entry *l, *devices;
   gboolean on = FALSE;
   gboolean ok;
   guint guards;
@@ -851,30 +849,14 @@ bolt_power_initialize (GInitable    *initable,
                            (GCallback) handle_uevent_udev,
                            power, 0);
 
-  e = bolt_udev_new_enumerate (power->udev, NULL);
-  udev_enumerate_add_match_subsystem (e, "wmi");
-  udev_enumerate_add_match_property (e, "DRIVER", "intel-wmi-thunderbolt");
-
-  udev_enumerate_scan_devices (e);
-  devices = udev_enumerate_get_list_entry (e);
-
-  udev_list_entry_foreach (l, devices)
+  ok = bolt_udev_detect_force_power (power->udev, &power->path, &err);
+  if (!ok)
     {
-      g_autofree char *path = NULL;
-      const char *syspath;
-
-      syspath = udev_list_entry_get_name (l);
-      path = g_build_filename (syspath, "force_power", NULL);
-
-      if (g_file_test (path, G_FILE_TEST_IS_REGULAR))
-        {
-          power->path = g_steal_pointer (&path);
-          break;
-        }
+      bolt_warn_err (err, LOG_TOPIC ("power"),
+                     "failure while detecting force power");
+      g_clear_error (&err);
     }
 
-  udev_enumerate_unref (e);
-
   bolt_msg (LOG_TOPIC ("power"), "force power support: %s",
             bolt_yesno (power->path != NULL));
 
@@ -998,6 +980,11 @@ bolt_power_wait_timeout (gpointer user_data)
   BoltPower *power = user_data;
   gboolean ok;
 
+  if (power->path == NULL)
+    /* force power support got removed while being used,
+    * this was already complained about, so ignore it */
+    return G_SOURCE_REMOVE;
+
   /* we just removed the last active guard */
   ok = bolt_power_switch_toggle (power, FALSE, &err);
 
@@ -1075,44 +1062,54 @@ handle_uevent_wmi (BoltPower          *power,
                    const char         *action,
                    struct udev_device *device)
 {
+  g_autoptr(GError) err = NULL;
   g_autofree char *path = NULL;
   const char *name;
   const char *syspath;
   gboolean changed = FALSE;
+  gboolean ok;
 
   syspath = udev_device_get_syspath (device);
   name = udev_device_get_sysname (device);
 
-  bolt_debug (LOG_TOPIC ("power"), "uevent: wmi %s [%s %s]",
-              name, syspath, power->path ? : "<unset>");
+  bolt_debug (LOG_TOPIC ("power"), "uevent: wmi %s %s [%s %s]",
+              action, name, syspath, power->path ? : "<unset>");
 
-  path = g_build_filename (syspath, "force_power", NULL);
+  if (!bolt_streq (action, "change"))
+    return;
 
-  if (bolt_streq (action, "bind") &&
-      bolt_streq (name, INTEL_WMI_THUNDERBOLT_GUID))
+  ok = bolt_udev_detect_force_power (power->udev, &path, &err);
+
+  if (!ok)
     {
+      bolt_warn_err (err, LOG_TOPIC ("udev"),
+                     "failed to detect force_power support");
+      g_clear_error (&err);
+    }
 
-      if (g_file_test (path, G_FILE_TEST_IS_REGULAR))
+  if (path != NULL)
+    {
+      if (power->path == NULL || !bolt_streq (path, power->path))
         {
           bolt_set_str (&power->path, g_steal_pointer (&path));
-          power->state = BOLT_FORCE_POWER_UNSET;
           changed = TRUE;
         }
     }
-  else if (bolt_streq (action, "unbind") &&
-           bolt_streq (path, power->path))
+  else if (path == NULL && power->path != NULL)
     {
       if (power->state > BOLT_FORCE_POWER_OFF)
         bolt_warn (LOG_TOPIC ("power"),
                    "force power supported removed while active");
 
       g_clear_pointer (&power->path, g_free);
-      power->state = BOLT_FORCE_POWER_UNSET;
       changed = TRUE;
     }
 
   if (changed)
     {
+      /* if changed, we don't know our current state */
+      power->state = BOLT_FORCE_POWER_UNSET;
+
       g_object_notify_by_pspec (G_OBJECT (power),
                                 power_props[PROP_STATE]);
       g_object_notify_by_pspec (G_OBJECT (power),
@@ -1167,13 +1164,7 @@ bolt_power_switch_toggle (BoltPower *power,
   int fd;
 
   g_return_val_if_fail (BOLT_IS_POWER (power), FALSE);
-
-  if (power->path == NULL)
-    {
-      g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
-                           "force power not supported");
-      return FALSE;
-    }
+  g_return_val_if_fail (power->path != NULL, FALSE);
 
   bolt_info (LOG_TOPIC ("power"), "setting force_power to %s",
              on ? "ON" : "OFF");
@@ -1332,6 +1323,7 @@ handle_force_power (BoltExported          *object,
 
   g_variant_get (params, "(&s&s)", &who, &flags);
 
+  /* TODO: log errors */
   guard = bolt_power_acquire_full (power, who, (pid_t) pid, error);
   if (guard == NULL)
     return NULL;
@@ -1434,6 +1426,14 @@ bolt_power_acquire_full (BoltPower  *power,
 
   g_return_val_if_fail (BOLT_IS_POWER (power), NULL);
   g_return_val_if_fail (who != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
+  if (power->path == NULL)
+    {
+      g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
+                           "force power not supported");
+      return FALSE;
+    }
 
   id = bolt_power_gen_guard_id (power, error);
 
@@ -1454,9 +1454,7 @@ bolt_power_acquire_full (BoltPower  *power,
 
       if (!ok)
         {
-          bolt_warn_err (err, LOG_TOPIC ("power"),
-                         "failed to toggle force_power");
-          g_propagate_error (error, g_steal_pointer (&err));
+          bolt_error_propagate (error, &err);
           return NULL;
         }
     }
