@@ -39,6 +39,7 @@ struct _BoltStore
   GObject object;
 
   GFile  *root;
+  GFile  *domains;
   GFile  *devices;
   GFile  *keys;
   GFile  *times;
@@ -75,15 +76,10 @@ bolt_store_finalize (GObject *object)
 {
   BoltStore *store = BOLT_STORE (object);
 
-  if (store->root)
-    g_clear_object (&store->root);
-
-  if (store->devices)
-    g_clear_object (&store->devices);
-
-  if (store->keys)
-    g_clear_object (&store->keys);
-
+  g_clear_object (&store->root);
+  g_clear_object (&store->domains);
+  g_clear_object (&store->devices);
+  g_clear_object (&store->keys);
   g_clear_object (&store->times);
 
   G_OBJECT_CLASS (bolt_store_parent_class)->finalize (object);
@@ -138,6 +134,7 @@ bolt_store_constructed (GObject *obj)
   BoltStore *store = BOLT_STORE (obj);
 
   store->devices = g_file_get_child (store->root, "devices");
+  store->domains = g_file_get_child (store->root, "domains");
   store->keys = g_file_get_child (store->root, "keys");
   store->times = g_file_get_child (store->root, "times");
 }
@@ -187,7 +184,7 @@ bolt_store_class_init (BoltStoreClass *klass)
 }
 
 /* internal methods */
-
+#define DOMAIN_GROUP "domain"
 #define DEVICE_GROUP "device"
 #define USER_GROUP "user"
 
@@ -219,7 +216,8 @@ bolt_store_config_load (BoltStore *store,
   gboolean ok;
   gsize len;
 
-  g_return_val_if_fail (store != NULL, FALSE);
+  g_return_val_if_fail (store != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
   sf = g_file_get_child (store->root, CFG_FILE);
   ok = g_file_load_contents (sf, NULL,
@@ -249,6 +247,10 @@ bolt_store_config_save (BoltStore *store,
   gboolean ok;
   gsize len;
 
+  g_return_val_if_fail (store != NULL, FALSE);
+  g_return_val_if_fail (config != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
   sf = g_file_get_child (store->root, CFG_FILE);
   data = g_key_file_to_data (config, &len, error);
 
@@ -266,8 +268,9 @@ bolt_store_config_save (BoltStore *store,
 }
 
 GStrv
-bolt_store_list_uids (BoltStore *store,
-                      GError   **error)
+bolt_store_list_uids (BoltStore  *store,
+                      const char *type,
+                      GError    **error)
 {
   g_autoptr(GError) err = NULL;
   g_autoptr(GDir) dir   = NULL;
@@ -275,16 +278,31 @@ bolt_store_list_uids (BoltStore *store,
   g_autoptr(GPtrArray) ids = NULL;
   const char *name;
 
+  g_return_val_if_fail (BOLT_IS_STORE (store), NULL);
+  g_return_val_if_fail (type != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
+  if (bolt_streq (type, "devices"))
+    path = g_file_get_path (store->devices);
+  if (bolt_streq (type, "domains"))
+    path = g_file_get_path (store->domains);
+
+  if (path == NULL)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+                   "unknown stored typed '%s'", type);
+      return NULL;
+    }
+
   ids = g_ptr_array_new ();
 
-  path = g_file_get_path (store->devices);
   dir = g_dir_open (path, 0, &err);
   if (dir == NULL)
     {
       if (bolt_err_notfound (err))
         return bolt_strv_from_ptr_array (&ids);
 
-      g_propagate_error (error, g_steal_pointer (&err));
+      bolt_error_propagate (error, &err);
       return NULL;
     }
 
@@ -299,6 +317,150 @@ bolt_store_list_uids (BoltStore *store,
   return bolt_strv_from_ptr_array (&ids);
 }
 
+gboolean
+bolt_store_put_domain (BoltStore  *store,
+                       BoltDomain *domain,
+                       GError    **error)
+{
+  g_autoptr(GError) err = NULL;
+  g_autoptr(GFile) entry = NULL;
+  g_autoptr(GKeyFile) kf = NULL;
+  g_autofree char *path = NULL;
+  char * const * bootacl = NULL;
+  const char *uid;
+  gboolean ok;
+  gsize len;
+
+  g_return_val_if_fail (BOLT_IS_STORE (store), FALSE);
+  g_return_val_if_fail (BOLT_IS_DOMAIN (domain), FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  uid = bolt_domain_get_uid (domain);
+  g_assert (uid);
+
+  entry = g_file_get_child (store->domains, uid);
+
+  ok = bolt_fs_make_parent_dirs (entry, error);
+  if (!ok)
+    return FALSE;
+
+  kf = g_key_file_new ();
+  path = g_file_get_path (entry);
+  ok = g_key_file_load_from_file (kf,
+                                  path,
+                                  G_KEY_FILE_KEEP_COMMENTS,
+                                  &err);
+
+  if (!ok && !bolt_err_notfound (err))
+    {
+      bolt_warn_err (err, LOG_TOPIC ("store"),
+                     "error loading existing domain");
+      g_clear_error (&err);
+      /* not fatal, keep going */
+    }
+
+  bootacl = bolt_domain_get_bootacl (domain);
+  len = bolt_strv_length (bootacl);
+
+  g_key_file_set_string_list (kf,
+                              DOMAIN_GROUP,
+                              "bootacl",
+                              (const char * const *) bootacl,
+                              len);
+
+  ok = g_key_file_save_to_file (kf, path, error);
+
+  if (!ok)
+    return FALSE;
+
+  g_object_set (G_OBJECT (domain),
+                "store", store,
+                NULL);
+
+  return ok;
+}
+
+BoltDomain *
+bolt_store_get_domain (BoltStore  *store,
+                       const char *uid,
+                       GError    **error)
+{
+  g_autoptr(GKeyFile) kf = NULL;
+  g_autoptr(GFile) db = NULL;
+  g_autoptr(GError) err = NULL;
+  g_autofree char *path = NULL;
+  g_auto(GStrv) bootacl = NULL;
+  BoltDomain *domain = NULL;
+  gboolean ok;
+
+  g_return_val_if_fail (store != NULL, NULL);
+  g_return_val_if_fail (uid != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
+  db = g_file_get_child (store->domains, uid);
+  path = g_file_get_path (db);
+
+  kf = g_key_file_new ();
+  ok = g_key_file_load_from_file (kf, path, G_KEY_FILE_NONE, error);
+
+  if (!ok)
+    return NULL;
+
+  bootacl = g_key_file_get_string_list (kf,
+                                        DOMAIN_GROUP,
+                                        "bootacl",
+                                        NULL,
+                                        &err);
+
+  if (bootacl == NULL && !bolt_err_notfound (err))
+    {
+      bolt_warn_err (err, LOG_DOM_UID (uid), LOG_TOPIC ("store"),
+                     "failed to parse bootacl for domain '%s'",
+                     uid);
+
+      g_clear_error (&err);
+    }
+
+  if (bolt_strv_isempty (bootacl))
+    g_clear_pointer (&bootacl, g_strfreev);
+
+  domain = g_object_new (BOLT_TYPE_DOMAIN,
+                         "store", store,
+                         "uid", uid,
+                         "bootacl", bootacl,
+                         NULL);
+
+  return domain;
+}
+
+gboolean
+bolt_store_del_domain (BoltStore  *store,
+                       BoltDomain *domain,
+                       GError    **error)
+{
+  g_autoptr(GFile) path = NULL;
+  const char *uid;
+  gboolean ok;
+
+  g_return_val_if_fail (store != NULL, FALSE);
+  g_return_val_if_fail (domain != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  uid = bolt_domain_get_uid (domain);
+
+  path = g_file_get_child (store->domains, uid);
+  ok = g_file_delete (path, NULL, error);
+
+  if (!ok)
+    return FALSE;
+
+  g_object_set (domain,
+                "store", NULL,
+                NULL);
+
+  return TRUE;
+}
+
 gboolean
 bolt_store_put_device (BoltStore  *store,
                        BoltDevice *device,
@@ -308,10 +470,13 @@ bolt_store_put_device (BoltStore  *store,
 {
   g_autoptr(GFile) entry = NULL;
   g_autoptr(GKeyFile) kf = NULL;
-  g_autofree char *data  = NULL;
+  g_autoptr(GError) err = NULL;
+  g_autofree char *data = NULL;
+  g_autofree char *path = NULL;
   BoltDeviceType type;
   const char *uid;
   const char *label;
+  gboolean fresh;
   gboolean ok;
   guint64 ctime;
   guint64 atime;
@@ -319,8 +484,10 @@ bolt_store_put_device (BoltStore  *store,
   gsize len;
   guint keystate = 0;
 
-  g_return_val_if_fail (store != NULL, FALSE);
-  g_return_val_if_fail (device != NULL, FALSE);
+  g_return_val_if_fail (BOLT_IS_STORE (store), FALSE);
+  g_return_val_if_fail (BOLT_IS_DEVICE (device), FALSE);
+  g_return_val_if_fail (key == NULL || BOLT_IS_KEY (key), FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
   uid = bolt_device_get_uid (device);
   g_assert (uid);
@@ -333,6 +500,17 @@ bolt_store_put_device (BoltStore  *store,
 
   kf = g_key_file_new ();
 
+  path = g_file_get_path (entry);
+
+  ok = g_key_file_load_from_file (kf,
+                                  path,
+                                  G_KEY_FILE_KEEP_COMMENTS,
+                                  &err);
+
+  if (!ok && bolt_err_exists (err))
+    bolt_warn_err (err, LOG_TOPIC ("store"), LOG_DEV_UID (uid),
+                   "could not load previously stored device");
+
   g_key_file_set_string (kf, DEVICE_GROUP, "name", bolt_device_get_name (device));
   g_key_file_set_string (kf, DEVICE_GROUP, "vendor", bolt_device_get_vendor (device));
 
@@ -364,7 +542,7 @@ bolt_store_put_device (BoltStore  *store,
 
   if (key)
     {
-      g_autoptr(GError) err  = NULL;
+      g_clear_error (&err);
 
       ok = bolt_store_put_key (store, uid, key, &err);
 
@@ -384,6 +562,8 @@ bolt_store_put_device (BoltStore  *store,
   if (!ok)
     return FALSE;
 
+  fresh = bolt_device_get_stored (device) == FALSE;
+
   g_object_set (device,
                 "store", store,
                 "policy", policy,
@@ -391,7 +571,8 @@ bolt_store_put_device (BoltStore  *store,
                 "storetime", stime,
                 NULL);
 
-  g_signal_emit (store, signals[SIGNAL_DEVICE_ADDED], 0, uid);
+  if (fresh)
+    g_signal_emit (store, signals[SIGNAL_DEVICE_ADDED], 0, uid);
 
   ctime = bolt_device_get_conntime (device);
   atime = bolt_device_get_authtime (device);
@@ -404,7 +585,9 @@ bolt_store_put_device (BoltStore  *store,
 }
 
 BoltDevice *
-bolt_store_get_device (BoltStore *store, const char *uid, GError **error)
+bolt_store_get_device (BoltStore  *store,
+                       const char *uid,
+                       GError    **error)
 {
   g_autoptr(GKeyFile) kf = NULL;
   g_autoptr(GFile) db = NULL;
@@ -424,8 +607,9 @@ bolt_store_get_device (BoltStore *store, const char *uid, GError **error)
   guint64 ctime = 0;
   gsize len;
 
-  g_return_val_if_fail (store != NULL, FALSE);
-  g_return_val_if_fail (uid != NULL, FALSE);
+  g_return_val_if_fail (BOLT_IS_STORE (store), NULL);
+  g_return_val_if_fail (uid != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
   db = g_file_get_child (store->devices, uid);
   ok = g_file_load_contents (db, NULL,
@@ -531,6 +715,10 @@ bolt_store_del_device (BoltStore  *store,
   g_autoptr(GFile) devpath = NULL;
   gboolean ok;
 
+  g_return_val_if_fail (BOLT_IS_STORE (store), FALSE);
+  g_return_val_if_fail (uid != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
   devpath = g_file_get_child (store->devices, uid);
   ok = g_file_delete (devpath, NULL, error);
 
@@ -599,8 +787,7 @@ bolt_store_get_times (BoltStore  *store,
         {
           if (error != NULL)
             {
-              g_propagate_error (error, g_steal_pointer (&err));
-              res = FALSE;
+              res = bolt_error_propagate (error, &err);
               break;
             }
 
@@ -683,8 +870,7 @@ bolt_store_put_times (BoltStore  *store,
         {
           if (error != NULL)
             {
-              g_propagate_error (error, g_steal_pointer (&err));
-              res = FALSE;
+              res = bolt_error_propagate (error, &err);
               break;
             }
 
@@ -708,6 +894,11 @@ bolt_store_del_time (BoltStore  *store,
   g_autofree char *name = NULL;
   gboolean ok;
 
+  g_return_val_if_fail (BOLT_IS_STORE (store), FALSE);
+  g_return_val_if_fail (uid != NULL, FALSE);
+  g_return_val_if_fail (timesel != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
   name = g_strdup_printf ("%s.%s", uid, timesel);
   pathfile = g_file_get_child (store->times, name);
   ok = g_file_delete (pathfile, NULL, error);
@@ -741,8 +932,7 @@ bolt_store_del_times (BoltStore  *store,
         {
           if (error != NULL)
             {
-              g_propagate_error (error, g_steal_pointer (&err));
-              res = FALSE;
+              res = bolt_error_propagate (error, &err);
               break;
             }
 
@@ -766,6 +956,11 @@ bolt_store_put_key (BoltStore  *store,
   g_autoptr(GFile) keypath = NULL;
   gboolean ok;
 
+  g_return_val_if_fail (BOLT_IS_STORE (store), FALSE);
+  g_return_val_if_fail (uid != NULL, FALSE);
+  g_return_val_if_fail (BOLT_IS_KEY (key), FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
   keypath = g_file_get_child (store->keys, uid);
   ok = bolt_fs_make_parent_dirs (keypath, error);
 
@@ -784,6 +979,9 @@ bolt_store_have_key (BoltStore  *store,
   g_autoptr(GError) err = NULL;
   guint key = BOLT_KEY_MISSING;
 
+  g_return_val_if_fail (BOLT_IS_STORE (store), key);
+  g_return_val_if_fail (uid != NULL, key);
+
   keypath = g_file_get_child (store->keys, uid);
   keyinfo = g_file_query_info (keypath, "standard::*", 0, NULL, &err);
 
@@ -802,6 +1000,10 @@ bolt_store_get_key (BoltStore  *store,
 {
   g_autoptr(GFile) keypath = NULL;
 
+  g_return_val_if_fail (BOLT_IS_STORE (store), NULL);
+  g_return_val_if_fail (uid != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
   keypath = g_file_get_child (store->keys, uid);
 
   return bolt_key_load_file (keypath, error);
@@ -815,6 +1017,10 @@ bolt_store_del_key (BoltStore  *store,
   g_autoptr(GFile) keypath = NULL;
   gboolean ok;
 
+  g_return_val_if_fail (BOLT_IS_STORE (store), FALSE);
+  g_return_val_if_fail (uid != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
   keypath = g_file_get_child (store->keys, uid);
   ok = g_file_delete (keypath, NULL, error);
 
@@ -830,8 +1036,9 @@ bolt_store_del (BoltStore  *store,
   const char *uid;
   gboolean ok;
 
-  g_return_val_if_fail (store != NULL, FALSE);
-  g_return_val_if_fail (dev != NULL, FALSE);
+  g_return_val_if_fail (BOLT_IS_STORE (store), FALSE);
+  g_return_val_if_fail (BOLT_IS_DEVICE (dev), FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
   uid = bolt_device_get_uid (dev);
 
@@ -861,3 +1068,24 @@ bolt_store_del (BoltStore  *store,
 
   return ok;
 }
+
+BoltJournal *
+bolt_store_open_journal (BoltStore  *store,
+                         const char *type,
+                         const char *name,
+                         GError    **error)
+{
+  g_autoptr(GFile) root = NULL;
+  BoltJournal *journal;
+
+  g_return_val_if_fail (store != NULL, NULL);
+  g_return_val_if_fail (type != NULL, NULL);
+  g_return_val_if_fail (name != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
+  root = g_file_get_child (store->root, type);
+
+  journal = bolt_journal_new (root, name, error);
+
+  return journal;
+}
