@@ -7,6 +7,10 @@ export PYTHONPATH="/usr/share/glib-2.0"
 
 rm -rf /build
 mkdir /build
-meson . /build
+meson -Db_coverage=true . /build
 ninja -C /build
 meson test -C /build --verbose
+
+if [[ -x "$(command -v lcov)" ]]; then
+    ninja -C /build coverage
+fi
