@@ -38,6 +38,7 @@
 #include <stdlib.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#include <stdlib.h>
 
 typedef struct udev_device udev_device;
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (udev_device, udev_device_unref);
@@ -111,25 +112,48 @@ test_power_basic (TestPower *tt, gconstpointer user)
 {
   g_autoptr(BoltPower) power = NULL;
   g_autoptr(GError) err = NULL;
+  g_autoptr(BoltUdev) udev = NULL;
+  g_autoptr(GFile) statedir = NULL;
   g_autoptr(BoltPowerGuard) guard = NULL;
+  g_autoptr(BoltPower) guard_power = NULL;
   g_autofree char *guard_id = NULL;
   g_autofree char *guard_who = NULL;
+  g_autofree char *guard_path = NULL;
+  g_autofree char *guard_fifo = NULL;
+  g_autofree char *rundir = NULL;
+  gulong guard_pid;
   BoltPowerState state;
   gboolean supported;
   gboolean on;
   const char *fp;
+  guint timeout;
 
   power = make_bolt_power_timeout (tt, 0);
 
   g_object_get (power,
+                "rundir", &rundir,
+                "statedir", &statedir,
+                "udev", &udev,
                 "supported", &supported,
                 "state", &state,
+                "timeout", &timeout,
                 NULL);
 
+  g_assert_nonnull (rundir);
+  g_assert_nonnull (statedir);
+  g_assert_nonnull (udev);
+  g_assert (udev == tt->udev);
   g_assert_false (supported);
   g_assert (state == BOLT_FORCE_POWER_UNSET);
   state = bolt_power_get_state (power);
   g_assert (state == BOLT_FORCE_POWER_UNSET);
+  g_assert_cmpuint (timeout, ==, 0);
+
+  /* force power is unsupported, check the error handling */
+  guard = bolt_power_acquire (power, &err);
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED);
+  g_assert_null (guard);
+  g_clear_pointer (&err, g_error_free);
 
   /* add the force power sysfs device,
    * now it must be reported as supported */
@@ -171,13 +195,24 @@ test_power_basic (TestPower *tt, gconstpointer user)
   g_object_get (guard,
                 "id", &guard_id,
                 "who", &guard_who,
+                "power", &guard_power,
+                "path", &guard_path,
+                "pid", &guard_pid,
+                "fifo", &guard_fifo,
                 NULL);
 
   g_assert_nonnull (guard_id);
   g_assert_nonnull (guard_who);
+  g_assert_nonnull (guard_power);
+  g_assert_nonnull (guard_path);
 
   g_assert_cmpstr (guard_id, ==, "1");
   g_assert_cmpstr (guard_who, ==, "boltd");
+  g_assert_cmpuint (guard_pid, ==, getpid ());
+
+  g_assert_cmpstr (guard_id, ==, bolt_power_guard_get_id (guard));
+  g_assert_cmpstr (guard_who, ==, bolt_power_guard_get_who (guard));
+  g_assert_cmpuint (guard_pid, ==, bolt_power_guard_get_pid (guard));
 
   /* set of OFF */
   g_clear_object (&guard);
@@ -601,7 +636,7 @@ test_power_guards_fifo (TestPower *tt, gconstpointer user)
   tid = g_timeout_add_seconds (5, on_timeout_warn_quit_loop, loop);
 
   /* schedule a closing of the fifo */
-  g_idle_add (on_cb_close_fd, (gpointer) &fd);
+  g_idle_add (on_cb_close_fd, (gpointer) & fd);
 
   g_signal_connect (power, "notify::state",
                     G_CALLBACK (on_notify_quit_loop),
